@@ -3,15 +3,17 @@
 //! A wrapper around [git(1)](https://git-scm.com/docs/git) inspired by
 //! [`GitPython`](https://github.com/gitpython-developers/GitPython).
 
-#![allow(unknown_lints)]
 #![warn(clippy::all)]
 
-pub use posix_errors::PosixError;
+pub use posix_errors::{PosixError, EACCES, EINVAL, ENOENT};
 use std::collections::HashMap;
 use std::path::{Path, PathBuf};
 use std::process::Command;
 use std::process::Output;
 
+/// Experimental stuff
+pub mod x;
+
 macro_rules! cmd {
     ($args:expr) => {
         Command::new("git").args($args).output()
@@ -26,6 +28,7 @@ macro_rules! cmd {
 /// # Errors
 ///
 /// Will return [`PosixError`] if command exits with an error code.
+#[inline]
 pub fn ls_remote(args: &[&str]) -> Result<Output, PosixError> {
     let result = cmd!("ls-remote", args);
 
@@ -41,6 +44,7 @@ pub fn ls_remote(args: &[&str]) -> Result<Output, PosixError> {
 /// # Errors
 ///
 /// Will return [`PosixError`] if command exits with an error code.
+#[inline]
 pub fn tags_from_remote(url: &str) -> Result<Vec<String>, PosixError> {
     let mut vec = Vec::new();
     let output = ls_remote(&["--refs", "--tags", url])?;
@@ -61,6 +65,9 @@ pub fn tags_from_remote(url: &str) -> Result<Vec<String>, PosixError> {
     }
 }
 
+/// Failed to change configuration file
+#[allow(missing_docs)]
+#[derive(Debug)]
 pub enum ConfigSetError {
     InvalidSectionOrKey(String),
     InvalidConfigFile(String),
@@ -74,6 +81,7 @@ pub enum ConfigSetError {
 /// # Panics
 ///
 /// When git-config(1) execution fails
+#[inline]
 pub fn config_file_set(file: &Path, key: &str, value: &str) -> Result<(), ConfigSetError> {
     let args = &["--file", file.to_str().expect("UTF-8 encoding"), key, value];
     let mut cmd = Command::new("git");
@@ -105,6 +113,7 @@ pub fn config_file_set(file: &Path, key: &str, value: &str) -> Result<(), Config
 ///
 /// Will return [`PosixError`] if command exits with an error code.
 /// TODO Return a custom error type
+#[inline]
 pub fn resolve_head(remote: &str) -> Result<String, PosixError> {
     let proc =
         cmd!("ls-remote", vec!["--symref", remote, "HEAD"]).expect("Failed to execute git command");
@@ -122,7 +131,7 @@ pub fn resolve_head(remote: &str) -> Result<String, PosixError> {
         return Ok(split
             .next()
             .expect("Failed to parse default branch")
-            .to_string());
+            .to_owned());
     }
 
     Err(PosixError::from(proc))
@@ -140,6 +149,7 @@ struct RemoteLine {
 }
 
 /// Represents a git remote
+#[allow(missing_docs)]
 #[derive(Clone, Debug, Eq, PartialEq)]
 pub struct Remote {
     pub name: String,
@@ -160,7 +170,7 @@ fn cwd() -> Result<PathBuf, RepoError> {
 pub struct AbsoluteDirPath(PathBuf);
 
 impl AbsoluteDirPath {
-    fn new(path: &Path) -> Result<AbsoluteDirPath, RepoError> {
+    fn new(path: &Path) -> Result<Self, RepoError> {
         let path_buf;
         if path.is_absolute() {
             path_buf = path.to_path_buf();
@@ -170,7 +180,7 @@ impl AbsoluteDirPath {
             return Err(RepoError::AbsolutionError(path.to_path_buf()));
         }
 
-        Ok(AbsoluteDirPath(path_buf))
+        Ok(Self(path_buf))
     }
 }
 
@@ -178,26 +188,48 @@ impl AbsoluteDirPath {
 ///
 /// This wrapper allows to keep track of optional *git-dir* and *work-tree* directories when
 /// executing commands. This functionality was needed for `glv` & `git-stree` project.
-
 #[derive(Clone, Debug)]
 pub enum Repository {
+    /// Bare repository *without* work-tree
     Bare {
+        /// GIT_DIR
         git_dir: AbsoluteDirPath,
     },
+    /// Bare repository *with* work-tree
     Normal {
+        /// GIT_DIR
         git_dir: AbsoluteDirPath,
+        /// WORK_TREE
         work_tree: AbsoluteDirPath,
     },
 }
 
-#[derive(Debug, PartialEq)]
+/// Error during repository instantiation
+#[allow(missing_docs)]
+#[derive(thiserror::Error, Debug, PartialEq)]
 pub enum RepoError {
+    #[error("GIT_DIR Not found")]
     GitDirNotFound,
+    #[error("Invalid directory: `{0}`")]
     InvalidDirectory(PathBuf),
+    #[error("Failed to canonicalize the path buffer: `{0}`")]
     AbsolutionError(PathBuf),
+    #[error("Failed to access current working directory")]
     FailAccessCwd,
 }
 
+impl From<RepoError> for PosixError {
+    #[inline]
+    fn from(e: RepoError) -> Self {
+        let msg = format!("{}", e);
+        match e {
+            RepoError::GitDirNotFound | RepoError::InvalidDirectory(_) => Self::new(ENOENT, msg),
+            RepoError::AbsolutionError(_) => Self::new(EINVAL, msg),
+            RepoError::FailAccessCwd => Self::new(EACCES, msg),
+        }
+    }
+}
+
 fn search_git_dir(start: &Path) -> Result<AbsoluteDirPath, RepoError> {
     let path;
     if start.is_absolute() {
@@ -253,13 +285,17 @@ fn git_dir_from_work_tree(work_tree: &AbsoluteDirPath) -> Result<AbsoluteDirPath
     AbsoluteDirPath::new(result.as_ref())
 }
 
+/// Invalid git reference was provided
+#[allow(missing_docs)]
 #[derive(Debug, PartialEq)]
 pub struct InvalidRefError;
 
 /// Getters
 impl Repository {
+    /// Return true if is bare repository
     #[must_use]
-    pub fn is_bare(&self) -> bool {
+    #[inline]
+    pub const fn is_bare(&self) -> bool {
         matches!(self, Self::Bare { .. })
     }
 
@@ -267,17 +303,19 @@ impl Repository {
     ///
     /// Panics of executing git-diff(1) fails
     #[must_use]
+    #[inline]
     pub fn is_clean(&self) -> bool {
-        let output = self.git().args(&["diff", "--quiet"]).output().unwrap();
-        if !output.status.success() {
-            return false;
-        }
-
-        let output = self.git().args(&["rev-parse", "HEAD"]).output().unwrap();
+        let output = self
+            .git()
+            .args(&["diff", "--quiet", "HEAD"])
+            .output()
+            .expect("Failed to execute git-diff(1)");
         output.status.success()
     }
 
+    /// Returns a `HashMap` of git remotes
     #[must_use]
+    #[inline]
     pub fn remotes(&self) -> Option<HashMap<String, Remote>> {
         let args = &["remote", "-v"];
         let mut cmd = self.git();
@@ -286,10 +324,6 @@ impl Repository {
             .output()
             .expect("Failed to execute git-remote(1)");
         if !out.status.success() {
-            eprintln!(
-                "Failed to execute git-remote(1):\n{}",
-                String::from_utf8_lossy(&out.stderr).to_string()
-            );
             return None;
         }
 
@@ -298,10 +332,10 @@ impl Repository {
         let mut remote_lines: Vec<RemoteLine> = vec![];
         for line in text.lines() {
             let mut split = line.trim().split('\t');
-            let name = split.next().expect("Remote name").to_string();
+            let name = split.next().expect("Remote name").to_owned();
             let rest = split.next().expect("Remote rest");
             let mut rest_split = rest.split(' ');
-            let url = rest_split.next().expect("Remote url").to_string();
+            let url = rest_split.next().expect("Remote url").to_owned();
             let dir = if rest_split.next().expect("Remote direction") == "(fetch)" {
                 RemoteDir::Fetch
             } else {
@@ -311,13 +345,13 @@ impl Repository {
         }
         for remote_line in remote_lines {
             let mut remote = my_map.remove(&remote_line.name).unwrap_or(Remote {
-                name: remote_line.name.to_string(),
+                name: remote_line.name.clone(),
                 push: None,
                 fetch: None,
             });
             match remote_line.dir {
-                RemoteDir::Fetch => remote.fetch = Some(remote_line.url.to_string()),
-                RemoteDir::Push => remote.push = Some(remote_line.url.to_string()),
+                RemoteDir::Fetch => remote.fetch = Some(remote_line.url.clone()),
+                RemoteDir::Push => remote.push = Some(remote_line.url.clone()),
             }
             my_map.insert(remote_line.name.clone(), remote);
         }
@@ -325,23 +359,27 @@ impl Repository {
         Some(my_map)
     }
 
+    /// Returns the HEAD commit id if ref HEAD exists
     // TODO return a Result with custom error type
     #[must_use]
+    #[inline]
     pub fn head(&self) -> Option<String> {
         let args = &["rev-parse", "HEAD"];
         let mut cmd = self.git();
         let out = cmd
             .args(args)
             .output()
-            .expect("Failed to execute git-commit(1)");
+            .expect("Failed to execute git-rev-parse(1)");
         if !out.status.success() {
             return None;
         }
-        let result = String::from_utf8_lossy(&out.stdout).to_string();
+        let result = String::from_utf8_lossy(&out.stdout).trim().to_owned();
         Some(result)
     }
 
+    /// Return path to git `WORK_TREE`
     #[must_use]
+    #[inline]
     pub fn work_tree(&self) -> Option<PathBuf> {
         match self {
             Self::Normal { work_tree, .. } => Some(work_tree.0.clone()),
@@ -351,18 +389,19 @@ impl Repository {
 
     /// Return true if the repo is sparse
     #[must_use]
+    #[inline]
     pub fn is_sparse(&self) -> bool {
         let path = self.git_dir_path().join("info").join("sparse-checkout");
         path.exists()
     }
 
-    fn git_dir(&self) -> &AbsoluteDirPath {
+    const fn git_dir(&self) -> &AbsoluteDirPath {
         match self {
             Self::Normal { git_dir, .. } | Self::Bare { git_dir } => git_dir,
         }
     }
 
-    fn git_dir_path(&self) -> &PathBuf {
+    const fn git_dir_path(&self) -> &PathBuf {
         match self {
             Self::Normal { git_dir, .. } | Self::Bare { git_dir } => &git_dir.0,
         }
@@ -371,6 +410,7 @@ impl Repository {
     /// # Errors
     ///
     /// Will return [`InvalidRefError`] if invalid reference provided
+    #[inline]
     pub fn short_ref(&self, long_ref: &str) -> Result<String, InvalidRefError> {
         let args = vec!["rev-parse", "--short", long_ref];
         let mut cmd = self.git();
@@ -392,6 +432,7 @@ impl Repository {
     /// # Errors
     ///
     /// Will return [`RepoError`] when fails to find repository
+    #[inline]
     pub fn discover(path: &Path) -> Result<Self, RepoError> {
         let git_dir = search_git_dir(path)?;
         if let Some(work_tree) = work_tree_from_git_dir(&git_dir)? {
@@ -403,12 +444,15 @@ impl Repository {
     /// # Errors
     ///
     /// Will return [`RepoError`] when fails to find repository
+    #[inline]
     pub fn default() -> Result<Self, RepoError> {
-        Repository::from_args(None, None, None)
+        Self::from_args(None, None, None)
     }
 
     #[must_use]
-    pub fn new(git_dir: AbsoluteDirPath, work_tree: Option<AbsoluteDirPath>) -> Self {
+    #[inline]
+    #[allow(clippy::shadow_reuse, clippy::missing_const_for_fn)]
+    fn new(git_dir: AbsoluteDirPath, work_tree: Option<AbsoluteDirPath>) -> Self {
         match work_tree {
             Some(work_tree) => Self::Normal { git_dir, work_tree },
             None => Self::Bare { git_dir },
@@ -422,6 +466,7 @@ impl Repository {
     /// # Errors
     ///
     /// Returns a string output when something goes horrible wrong
+    #[inline]
     pub fn create(path: &Path) -> Result<Self, String> {
         let mut cmd = Command::new("git");
         let out = cmd.arg("init").current_dir(&path).output().unwrap();
@@ -429,7 +474,7 @@ impl Repository {
         if out.status.success() {
             let work_tree = AbsoluteDirPath::new(path).unwrap();
             let git_dir = AbsoluteDirPath::new(&path.join(".git")).unwrap();
-            Ok(Repository::Normal { git_dir, work_tree })
+            Ok(Self::Normal { git_dir, work_tree })
         } else {
             Err(String::from_utf8_lossy(&out.stderr).to_string())
         }
@@ -442,6 +487,7 @@ impl Repository {
     /// # Errors
     ///
     /// Returns a string output when something goes horrible wrong
+    #[inline]
     pub fn create_bare(path: &Path) -> Result<Self, String> {
         let mut cmd = Command::new("git");
         let out = cmd
@@ -453,7 +499,7 @@ impl Repository {
 
         if out.status.success() {
             let git_dir = AbsoluteDirPath::new(path).unwrap();
-            Ok(Repository::Bare { git_dir })
+            Ok(Self::Bare { git_dir })
         } else {
             Err(String::from_utf8_lossy(&out.stderr).to_string())
         }
@@ -462,6 +508,7 @@ impl Repository {
     /// # Errors
     ///
     /// Will return [`RepoError`] when fails to find repository
+    #[inline]
     pub fn from_args(
         change: Option<&str>,
         git: Option<&str>,
@@ -481,7 +528,7 @@ impl Repository {
                     work_tree_from_git_dir(&git_dir)?
                 };
 
-                Ok(Repository::new(git_dir, work_tree))
+                Ok(Self::new(git_dir, work_tree))
             }
             (_, _, _) => {
                 let root = change.map_or_else(PathBuf::new, PathBuf::from);
@@ -489,7 +536,7 @@ impl Repository {
                     (Some(g_dir), None) => {
                         let git_dir = AbsoluteDirPath::new(&root.join(g_dir))?;
                         let work_tree = work_tree_from_git_dir(&git_dir)?;
-                        Ok(Repository::new(git_dir, work_tree))
+                        Ok(Self::new(git_dir, work_tree))
                     }
                     (None, Some(w_dir)) => {
                         let work_tree = AbsoluteDirPath::new(&root.join(w_dir))?;
@@ -504,34 +551,31 @@ impl Repository {
                     (None, None) => {
                         let git_dir = search_git_dir(&root)?;
                         let work_tree = work_tree_from_git_dir(&git_dir)?;
-                        Ok(Repository::new(git_dir, work_tree))
+                        Ok(Self::new(git_dir, work_tree))
                     }
                 }
             }
         }
     }
 
+    /// Returns a prepared git `Command` struct
     #[must_use]
-    pub fn git(&self) -> std::process::Command {
-        let mut cmd = std::process::Command::new("git");
+    #[inline]
+    pub fn git(&self) -> Command {
+        let mut cmd = Command::new("git");
         let git_dir = self.git_dir().0.to_str().expect("Convert to string");
         cmd.env("GIT_DIR", git_dir);
 
         if let Self::Normal { work_tree, .. } = self {
-            let work_tree = work_tree.0.to_str().expect("Convert to string");
-            cmd.env("GIT_WORK_TREE", work_tree);
-            cmd.current_dir(work_tree);
+            cmd.env("GIT_WORK_TREE", &work_tree.0);
+            cmd.current_dir(&work_tree.0);
         }
         cmd
     }
-
-    pub fn git_in_dir<P: AsRef<Path>>(&self, path: P) -> std::process::Command {
-        let mut cmd = self.git();
-        cmd.current_dir(path);
-        cmd
-    }
 }
 
+/// Failed to add subtree
+#[allow(missing_docs)]
 #[derive(Debug, PartialEq)]
 pub enum SubtreeAddError {
     BareRepository,
@@ -539,6 +583,8 @@ pub enum SubtreeAddError {
     Failure(String, i32),
 }
 
+/// Failed to pull changes from remote in to subtree
+#[allow(missing_docs)]
 #[derive(Debug, PartialEq)]
 pub enum SubtreePullError {
     BareRepository,
@@ -546,12 +592,16 @@ pub enum SubtreePullError {
     Failure(String, i32),
 }
 
+/// Failed to push changes from subtree to remote
+#[allow(missing_docs)]
 #[derive(Debug, PartialEq)]
 pub enum SubtreePushError {
     BareRepository,
     Failure(String, i32),
 }
 
+/// Failed to split subtree
+#[allow(missing_docs)]
 #[derive(Debug, PartialEq)]
 pub enum SubtreeSplitError {
     BareRepository,
@@ -559,28 +609,59 @@ pub enum SubtreeSplitError {
     Failure(String, i32),
 }
 
+/// Failed to read config
+#[allow(missing_docs)]
 #[derive(Debug)]
 pub enum ConfigReadError {
     InvalidSectionOrKey(String),
     InvalidConfigFile(String),
 }
 
-#[derive(Debug)]
+/// Failure to stage
+#[allow(missing_docs)]
+#[derive(thiserror::Error, Debug)]
 pub enum StagingError {
+    #[error("Bare repository")]
     BareRepository,
+    #[error("`{0}`")]
     Failure(String, i32),
+    #[error("File does not exist: `{0}`")]
     FileDoesNotExist(PathBuf),
-    InvalidPath(PathBuf),
-    UTF8Decode(PathBuf),
 }
 
-#[derive(Debug)]
+impl From<StagingError> for PosixError {
+    #[inline]
+    fn from(e: StagingError) -> Self {
+        let msg = format!("{}", e);
+        match e {
+            StagingError::BareRepository => Self::new(1, msg),
+            StagingError::FileDoesNotExist(_) => Self::new(ENOENT, msg),
+            StagingError::Failure(_, code) => Self::new(code, msg),
+        }
+    }
+}
+
+/// Error during stashing operation
+#[allow(missing_docs)]
+#[derive(thiserror::Error, Debug)]
+pub enum StashingError {
+    #[error("Failed to stash changes in GIT_WORK_TREE")]
+    Save(i32, String),
+    #[error("Failed to pop stashed changes in GIT_WORK_TREE")]
+    Pop(i32, String),
+}
+
+/// Error during committing
+#[allow(missing_docs)]
+#[derive(thiserror::Error, Debug)]
 pub enum CommitError {
+    #[error("`{0}`")]
     Failure(String, i32),
-    UTF8Decode(PathBuf),
+    #[error("Bare repository")]
     BareRepository,
 }
 
+/// Failed to find reference on remote
 #[derive(Debug)]
 pub enum RefSearchError {
     /// Thrown when `git-ls-remote(1)` fails to execute.
@@ -596,12 +677,14 @@ pub enum RefSearchError {
 }
 
 impl From<std::io::Error> for RefSearchError {
+    #[inline]
     fn from(prev: std::io::Error) -> Self {
         Self::IOError(prev)
     }
 }
 
 impl From<std::string::FromUtf8Error> for RefSearchError {
+    #[inline]
     fn from(prev: std::string::FromUtf8Error) -> Self {
         Self::UTF8Decode(prev)
     }
@@ -618,6 +701,7 @@ impl Repository {
     /// # Panics
     ///
     /// When `git-commit(1)` fails to execute
+    #[inline]
     pub fn commit(&self, message: &str) -> Result<(), CommitError> {
         if self.is_bare() {
             return Err(CommitError::BareRepository);
@@ -634,6 +718,42 @@ impl Repository {
         }
         Ok(())
     }
+
+    /// # Errors
+    ///
+    /// See [`CommitError`]
+    #[inline]
+    pub fn commit_extended(
+        &self,
+        message: &str,
+        allow_empty: bool,
+        no_verify: bool,
+    ) -> Result<(), CommitError> {
+        if self.is_bare() {
+            return Err(CommitError::BareRepository);
+        }
+
+        let mut cmd = self.git();
+        cmd.args(&["commit", "--quiet", "--no-edit"]);
+
+        if allow_empty {
+            cmd.arg("--allow-empty");
+        }
+
+        if no_verify {
+            cmd.arg("--no-verify");
+        }
+
+        cmd.args(&["--message", message]);
+
+        let out = cmd.output().expect("Failed to execute git-commit(1)");
+        if out.status.code().expect("Expected exit code") != 0 {
+            let msg = String::from_utf8_lossy(out.stderr.as_ref()).to_string();
+            let code = out.status.code().unwrap_or(1);
+            return Err(CommitError::Failure(msg, code));
+        }
+        Ok(())
+    }
     /// Return config value for specified key
     ///
     /// # Errors
@@ -643,6 +763,7 @@ impl Repository {
     /// # Panics
     ///
     /// Will panic if git exits with an unexpected error code. Expected codes are 0, 1 & 3.
+    #[inline]
     pub fn config(&self, key: &str) -> Result<String, ConfigReadError> {
         let out = self
             .git()
@@ -654,7 +775,7 @@ impl Repository {
             Ok(String::from_utf8(out.stdout)
                 .expect("UTF-8 encoding")
                 .trim()
-                .to_string())
+                .to_owned())
         } else {
             match out.status.code().unwrap() {
                 1 => Err(ConfigReadError::InvalidSectionOrKey(key.to_owned())),
@@ -679,6 +800,7 @@ impl Repository {
     /// # Errors
     ///
     /// When fails throws [`std::io::Error`]
+    #[inline]
     pub fn hack_read_file(&self, path: &Path) -> std::io::Result<Vec<u8>> {
         match self {
             Self::Normal { work_tree, .. } => {
@@ -686,14 +808,14 @@ impl Repository {
                 std::fs::read(absolute_path)
             }
             Self::Bare { .. } => {
-                let path = format!(":{}", path.to_str().expect("UTF-8 string"));
+                let file_path = format!(":{}", path.to_str().expect("UTF-8 string"));
                 let mut cmd = self.git();
-                cmd.args(&["show", &path]);
+                cmd.args(&["show", &file_path]);
                 let out = cmd.output().expect("Failed to execute git-show(1)");
                 if out.status.success() {
                     Ok(out.stdout)
                 } else if out.status.code().unwrap() == 128 {
-                    let msg = format!("Failed to read file: {:?}", path);
+                    let msg = format!("Failed to read file: {:?}", file_path);
                     Err(std::io::Error::new(std::io::ErrorKind::NotFound, msg))
                 } else {
                     let msg = String::from_utf8_lossy(out.stderr.as_ref());
@@ -703,7 +825,9 @@ impl Repository {
         }
     }
 
+    /// Returns true if the `first` commit is an ancestor of the `second` commit.
     #[must_use]
+    #[inline]
     pub fn is_ancestor(&self, first: &str, second: &str) -> bool {
         let args = vec!["merge-base", "--is-ancestor", first, second];
         let mut cmd = self.git();
@@ -715,6 +839,7 @@ impl Repository {
     /// # Errors
     ///
     /// See [`RefSearchError`]
+    #[inline]
     pub fn remote_ref_to_id(&self, remote: &str, git_ref: &str) -> Result<String, RefSearchError> {
         let proc = self.git().args(&["ls-remote", remote, git_ref]).output()?;
         if !proc.status.success() {
@@ -724,9 +849,9 @@ impl Repository {
         let stdout = String::from_utf8(proc.stdout)?;
         if let Some(first_line) = stdout.lines().next() {
             if let Some(id) = first_line.split('\t').next() {
-                return Ok(id.to_string());
+                return Ok(id.to_owned());
             }
-            return Err(RefSearchError::ParsingFailure(first_line.to_string()));
+            return Err(RefSearchError::ParsingFailure(first_line.to_owned()));
         }
 
         Err(RefSearchError::NotFound)
@@ -739,6 +864,7 @@ impl Repository {
     /// # Panics
     ///
     /// When git-sparse-checkout(1) execution fails
+    #[inline]
     pub fn sparse_checkout_add(&self, pattern: &str) -> Result<(), String> {
         let out = self
             .git()
@@ -761,26 +887,23 @@ impl Repository {
     /// # Panics
     ///
     /// Panics if fails to execute `git-add(1)`
+    #[inline]
     pub fn stage(&self, path: &Path) -> Result<(), StagingError> {
         if self.is_bare() {
             return Err(StagingError::BareRepository);
         }
 
-        let path = if path.is_absolute() {
+        let relative_path = if path.is_absolute() {
             path.strip_prefix(self.work_tree().unwrap()).unwrap()
         } else {
             path
         };
 
-        let file;
-        match path.to_str() {
-            Some(f) => file = f,
-            _ => return Err(StagingError::UTF8Decode(path.to_path_buf())),
-        }
-        let out = self.git().args(&["add", "--", file]).output().unwrap();
+        let file = relative_path.as_os_str();
+        let out = self.git().args(&["add", "--"]).arg(file).output().unwrap();
         match out.status.code().unwrap() {
             0 => Ok(()),
-            128 => Err(StagingError::FileDoesNotExist(path.to_path_buf())),
+            128 => Err(StagingError::FileDoesNotExist(relative_path.to_path_buf())),
             e => {
                 let msg = String::from_utf8_lossy(&out.stdout).to_string();
                 Err(StagingError::Failure(msg, e))
@@ -788,6 +911,48 @@ impl Repository {
         }
     }
 
+    /// Stash staged, unstaged and untracked files (keeps ignored files).
+    ///
+    /// # Errors
+    ///
+    /// See [`StashingError`]
+    #[inline]
+    pub fn stash_almost_all(&self, message: &str) -> Result<(), StashingError> {
+        let mut cmd = self.git();
+        cmd.arg("stash");
+        cmd.arg("--quiet");
+        cmd.args(&["--include-untracked", "-m", message]);
+
+        let out = cmd.output().expect("Failed to execute git-stash(1)");
+        if !out.status.success() {
+            let stderr = String::from_utf8_lossy(&out.stderr).to_string();
+            let code = out.status.code().unwrap_or(1);
+            return Err(StashingError::Save(code, stderr));
+        }
+        Ok(())
+    }
+
+    /// Pop stashed changes
+    ///
+    /// # Errors
+    ///
+    /// See [`StashingError`]
+    #[inline]
+    pub fn stash_pop(&self) -> Result<(), StashingError> {
+        let mut cmd = self.git();
+        let out = cmd
+            .args(&["stash", "pop", "--quiet", "--index"])
+            .output()
+            .expect("Failed to execute git-stash(1)");
+
+        if !out.status.success() {
+            let stderr = String::from_utf8_lossy(&out.stderr).to_string();
+            let code = out.status.code().unwrap_or(1);
+            return Err(StashingError::Pop(code, stderr));
+        }
+        Ok(())
+    }
+
     /// # Errors
     ///
     /// Fails if current repo is bare or dirty. In error cases see the provided string.
@@ -795,6 +960,7 @@ impl Repository {
     /// # Panics
     ///
     /// When git-subtree(1) execution fails
+    #[inline]
     pub fn subtree_add(
         &self,
         url: &str,
@@ -830,6 +996,7 @@ impl Repository {
     /// # Panics
     ///
     /// When git-subtree(1) execution fails
+    #[inline]
     pub fn subtree_split(&self, prefix: &str) -> Result<(), SubtreeSplitError> {
         if self.is_bare() {
             return Err(SubtreeSplitError::BareRepository);
@@ -842,14 +1009,25 @@ impl Repository {
         let args = vec!["-P", prefix, "--rejoin", "HEAD"];
         let mut cmd = self.git();
         cmd.arg("subtree").arg("split").args(args);
-        let out = cmd.output().expect("Failed to execute git-subtree(1)");
-
-        if out.status.success() {
-            Ok(())
-        } else {
-            let msg = String::from_utf8_lossy(out.stderr.as_ref()).to_string();
-            let code = out.status.code().unwrap_or(1);
-            Err(SubtreeSplitError::Failure(msg, code))
+        let result = cmd
+            .spawn()
+            .expect("Failed to execute git-subtree(1)")
+            .wait();
+        match result {
+            Ok(code) => {
+                if code.success() {
+                    Ok(())
+                } else {
+                    Err(SubtreeSplitError::Failure(
+                        "git-subtree split failed".to_owned(),
+                        1,
+                    ))
+                }
+            }
+            Err(e) => {
+                let msg = format!("{}", e);
+                Err(SubtreeSplitError::Failure(msg, 1))
+            }
         }
     }
 
@@ -860,6 +1038,7 @@ impl Repository {
     /// # Panics
     ///
     /// When git-subtree(1) execution fails
+    #[inline]
     pub fn subtree_pull(
         &self,
         remote: &str,
@@ -891,6 +1070,7 @@ impl Repository {
     /// # Errors
     ///
     /// Fails if current repo is bare. In other error cases see the provided message string.
+    #[inline]
     pub fn subtree_push(
         &self,
         remote: &str,
@@ -915,6 +1095,8 @@ impl Repository {
     }
 }
 
+/// Failed to resolve given value to a commit id
+#[allow(missing_docs)]
 #[derive(thiserror::Error, Debug)]
 pub enum InvalidCommitishError {
     #[error("Invalid reference or commit id: `{0}`")]
@@ -935,6 +1117,7 @@ impl Repository {
     /// # Panics
     ///
     /// When exit code of git-merge-base(1) is not 0 or 128
+    #[inline]
     pub fn merge_base(&self, ids: &[&str]) -> Result<Option<String>, InvalidCommitishError> {
         let output = self
             .git()
@@ -948,7 +1131,7 @@ impl Repository {
                 return Ok(None);
             }
             let result = tmp.trim_end();
-            return Ok(Some(result.to_string()));
+            return Ok(Some(result.to_owned()));
         }
         match output.status.code().expect("Getting status code") {
             128 => {
@@ -956,6 +1139,7 @@ impl Repository {
                 let e_ids = tmp.iter().map(ToString::to_string).collect();
                 Err(InvalidCommitishError::Multiple(e_ids))
             }
+            1 => Ok(None),
             code => {
                 panic!("Unexpected error code for merge-base: {}", code);
             }
@@ -964,33 +1148,15 @@ impl Repository {
 }
 
 #[cfg(test)]
-#[cfg(not(tarpaulin_include))]
 mod test {
 
-    // https://stackoverflow.com/a/63904992
-    macro_rules! function {
-        () => {{
-            fn f() {}
-            fn type_name_of<T>(_: T) -> &'static str {
-                std::any::type_name::<T>()
-            }
-            let name = type_name_of(f);
-
-            // Find and cut the rest of the path
-            match &name[..name.len() - 3].rfind(':') {
-                Some(pos) => &name[pos + 1..name.len() - 3],
-                None => &name[..name.len() - 3],
-            }
-        }};
-    }
-
     mod repository_initialization {
         use crate::{RepoError, Repository};
-        use tempdir::TempDir;
+        use tempfile::TempDir;
 
         #[test]
         fn git_dir_not_found() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let discovery_path = tmp_dir.path();
             let actual = Repository::discover(discovery_path);
             assert!(actual.is_err(), "Fail to find repo in an empty directory");
@@ -1001,7 +1167,7 @@ mod test {
 
         #[test]
         fn bare_repo() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create_bare(repo_path).unwrap();
             assert!(
@@ -1022,7 +1188,7 @@ mod test {
 
         #[test]
         fn normal_repo() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create(repo_path).unwrap();
 
@@ -1045,11 +1211,11 @@ mod test {
 
     mod is_bare {
         use crate::Repository;
-        use tempdir::TempDir;
+        use tempfile::TempDir;
 
         #[test]
         fn yes() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create_bare(repo_path).unwrap();
             assert!(repo.is_bare(), "Expected a bare repository");
@@ -1059,7 +1225,7 @@ mod test {
 
         #[test]
         fn no() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create(repo_path).unwrap();
             assert!(!repo.is_bare(), "Expected a non-bare repository");
@@ -1068,13 +1234,42 @@ mod test {
         }
     }
 
+    mod is_clean {
+        use crate::Repository;
+        use tempfile::TempDir;
+
+        #[test]
+        fn unstaged() {
+            let tmp_dir = TempDir::new().unwrap();
+            let repo_path = tmp_dir.path();
+            let repo = Repository::create(repo_path).unwrap();
+
+            let readme = repo_path.join("README.md");
+            std::fs::File::create(&readme).unwrap();
+            std::fs::write(&readme, "# README").unwrap();
+            assert!(!repo.is_clean(), "Repo is unclean if sth. is unstaged");
+        }
+
+        #[test]
+        fn staged() {
+            let tmp_dir = TempDir::new().unwrap();
+            let repo_path = tmp_dir.path();
+            let repo = Repository::create(repo_path).unwrap();
+
+            let readme = repo_path.join("README.md");
+            std::fs::File::create(&readme).unwrap();
+            repo.stage(&readme).unwrap();
+            assert!(!repo.is_clean(), "Repo is unclean if sth. is staged");
+        }
+    }
+
     mod config {
         use crate::Repository;
-        use tempdir::TempDir;
+        use tempfile::TempDir;
 
         #[test]
         fn config() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create_bare(repo_path).unwrap();
             let actual = repo.config("core.bare").unwrap();
@@ -1087,11 +1282,11 @@ mod test {
     mod sparse_checkout {
         use crate::Repository;
         use std::process::Command;
-        use tempdir::TempDir;
+        use tempfile::TempDir;
 
         #[test]
         fn is_sparse() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create(repo_path).unwrap();
             let mut cmd = Command::new("git");
@@ -1106,7 +1301,7 @@ mod test {
 
         #[test]
         fn not_sparse() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create(repo_path).unwrap();
             assert!(!repo.is_sparse(), "Not sparse repository")
@@ -1114,7 +1309,7 @@ mod test {
 
         #[test]
         fn add() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create(repo_path).unwrap();
             repo.git()
@@ -1130,11 +1325,11 @@ mod test {
 
     mod subtree_add {
         use crate::{Repository, SubtreeAddError};
-        use tempdir::TempDir;
+        use tempfile::TempDir;
 
         #[test]
         fn bare_repo() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create_bare(repo_path).unwrap();
             let actual =
@@ -1147,7 +1342,7 @@ mod test {
 
         #[test]
         fn dirty_work_tree() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create(repo_path).unwrap();
             let actual =
@@ -1160,7 +1355,7 @@ mod test {
 
         #[test]
         fn successfull() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create(repo_path).unwrap();
             let readme = repo_path.join("README.md");
@@ -1180,11 +1375,11 @@ mod test {
 
     mod subtree_pull {
         use crate::{Repository, SubtreePullError};
-        use tempdir::TempDir;
+        use tempfile::TempDir;
 
         #[test]
         fn bare_repo() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create_bare(repo_path).unwrap();
             let actual =
@@ -1197,7 +1392,7 @@ mod test {
 
         #[test]
         fn dirty_work_tree() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create(repo_path).unwrap();
             let actual =
@@ -1210,7 +1405,7 @@ mod test {
 
         #[test]
         fn successfull() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create(repo_path).unwrap();
             let readme = repo_path.join("README.md");
@@ -1239,11 +1434,11 @@ mod test {
     mod remote_ref_resolution {
         use crate::RefSearchError;
         use crate::Repository;
-        use tempdir::TempDir;
+        use tempfile::TempDir;
 
         #[test]
         fn not_found() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create(repo_path).unwrap();
             let result =
@@ -1255,7 +1450,7 @@ mod test {
 
         #[test]
         fn failure() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create(repo_path).unwrap();
             let result = repo.remote_ref_to_id("https://example.com/asd/foo", "v230.40.50");
@@ -1266,7 +1461,7 @@ mod test {
 
         #[test]
         fn successfull_search() {
-            let tmp_dir = TempDir::new(function!()).unwrap();
+            let tmp_dir = TempDir::new().unwrap();
             let repo_path = tmp_dir.path();
             let repo = Repository::create(repo_path).unwrap();
             let result = repo.remote_ref_to_id("https://github.com/kalkin/file-expert", "v0.9.0");
